import { Diagnostic } from '@codemirror/next/lint';
import { SyntaxNode, TreeCursor } from 'lezer-tree';
import log from 'loglevel';
import { Parser } from '../parser/parser';
import { gcodeSyntax } from '../gcode';

export class MoreHelpfulParser implements Parser {

    private diagnostics: Diagnostic[];

    private cursor;

    constructor() {}

    check(programText: string): Diagnostic[] {
        this.diagnostics = [];

        const start = new Date();
        const tree = gcodeSyntax.parser.parse(programText, {strict: false});

        tree.iterate({
            enter(type, from, to) {
                console.log(type.name);
            }
        });

        try {
            this.diagnose(tree.topNode);
        } catch (error) {
            if ('parse error' !== error.message) {
                throw error;
            }
        }

        console.log('Parse took: ' + (new Date().getMilliseconds() - start.getMilliseconds()) + 'ms');
        return this.getDiagnostics();
    }

    diagnose(node: SyntaxNode) {
        log.setLevel('debug');
        if (node.to === 0) {
            return;
        }
        this.cursor = node.cursor;
        do {
            log.debug(this.cursor.node.type.name);
        } while (this.cursor.next());
        log.debug('\n\n');
        log.debug(node.type.name);
        this.program(this.cursor);
    }

    private program(cursor: TreeCursor) {
        console.group(cursor.node.type.name);
        cursor.next();
        log.debug(cursor.node.type.name);
        this.program_start(cursor);
        cursor.next();
        log.debug(cursor.node.type.name);
        if ('Blocks' === cursor.node.type.name) {
            this.blocks(cursor);
        }
        log.debug(cursor.node.type.name);
        this.program_end(cursor);
        log.debug(cursor.node.type.name);
        console.groupEnd();
    }

    private program_start(cursor: TreeCursor) {
        console.group(cursor.node.type.name);
        cursor.firstChild();
        log.debug(cursor.node.type.name);
        MoreHelpfulParser.consumeEOB(cursor);
        this.expect('PERCENT', cursor);
        cursor.nextSibling();
        log.debug(cursor.node.type.name);
        MoreHelpfulParser.consumeComment(cursor);
        this.end_of_block(cursor);
        MoreHelpfulParser.consumeEOB(cursor);
        log.debug(cursor.node.type.name);
        this.expect('PROGRAM_NUMBER', cursor);
        cursor.next();
        log.debug(cursor.node.type.name);
        MoreHelpfulParser.consumeComment(cursor);
        this.end_of_block(cursor);
        log.debug(cursor.node.type.name);
        // cursor.next();
        console.groupEnd();
    }

    private blocks(cursor: TreeCursor) {
        console.group(cursor.node.type.name);
        cursor.next();
        // console.log(cursor.node.type.name);

        console.groupEnd();
    }

    private block(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private mblocks(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private commands(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private command(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private mcommands(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private end_of_block(cursor: TreeCursor) {
        console.group(cursor.node.type.name);
        cursor.next();
        log.debug(cursor.node.type.name);
        this.expect('EOB', cursor);
        cursor.next();
        console.groupEnd();
    }

    private program_end(cursor: TreeCursor) {
        console.group(cursor.node.type.name);
        cursor.next();
        this.expect('PERCENT', cursor);
        cursor.next();
        this.expect('End_Of_Block', cursor);
        MoreHelpfulParser.consumeEOB(cursor);
        console.groupEnd();
    }

    private if_statement(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private while_loop(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private do_loop(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private end_loop(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private assignment(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private blockskip(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private letter_command(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private goto_command(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private expression(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private subtracted_expression(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
        // console.log(cursor.node.type.name)
    }

    private multiplied_expression(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private modified_atom(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private negate(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private function(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    private atom(cursor: TreeCursor) {
        console.group(cursor.node.type.name);

        console.groupEnd();
    }

    getDiagnostics(): Diagnostic[] {
        return this.diagnostics;
    }

    private addDiagnostic(cursor: TreeCursor, msg: string): void {
        this.diagnostics.push({
            severity: 'error',
            message: msg,
            from: cursor.from ? cursor.from : 0,
            to: cursor.to ? cursor.to : cursor.to,
        });
    }

    private expect(expected: string, cursor: TreeCursor): void {
        if (expected !== cursor.node.type.name) {
            this.addDiagnostic(
              cursor,
              'got ' + cursor.node.type.name.toLowerCase() + ' expected ' + expected
            );
            // throw new Error('parse error');
        }
    }

    private static consumeEOB(cursor: TreeCursor): void {
        while ('End_Of_Block' === cursor.node.type.name || 'EOB' === cursor.node.type.name) {
            console.log(cursor.node.type.name);
            cursor.next();
        }
    }

    private static consumeComment(cursor: TreeCursor): void {
        while ('COMMENT' === cursor.node.type.name) {
            console.log(cursor.node.type.name);
            cursor.next();
        }
    }
}
